import logging
from collections import OrderedDict

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata
from typing import Optional
from urllib.parse import urljoin

from httpx import Client
from ipfabric_httpx_auth import PasswordCredentials, HeaderApiKey
from pydantic import BaseSettings

from ipfabric import models
from ipfabric.settings.user_mgmt import User


logger = logging.getLogger("ipfabric")

DEFAULT_ID = "$last"


class Settings(BaseSettings):
    ipf_url: str = ""
    ipf_version: str = ""
    ipf_token: str = ""
    ipf_verify: bool = True
    ipf_dev: bool = False
    ipf_username: str = ""
    ipf_password: str = ""

    class Config:
        env_file = ".env"
        env_file_encoding = "utf-8"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Needed for context"""
        pass


class IPFabricAPI(Client):
    def __init__(
        self,
        base_url: Optional[str] = None,
        api_version: Optional[str] = None,
        token: Optional[str] = None,
        snapshot_id: Optional[str] = DEFAULT_ID,
        username: Optional[str] = None,
        password: Optional[str] = None,
        **kwargs,
    ):
        """
        Initializes the IP Fabric Client
        :param base_url: str: IP Fabric instance provided in 'base_url' parameter, or the 'IPF_URL' environment variable
        :param token: str: API token or 'IPF_TOKEN' environment variable
        :param snapshot_id: str: IP Fabric snapshot ID to use by default for database actions - defaults to '$last'
        :param kwargs: dict: Keyword args to pass to httpx
        """
        with Settings() as settings:
            super().__init__(timeout=kwargs.get("timeout", True), headers={"Content-Type": "application/json"},
                             verify=kwargs.get("verify", settings.ipf_verify))
            base_url = base_url or settings.ipf_url
            if not base_url:
                raise RuntimeError("IP Fabric base_url not provided or IPF_URL not set")

            self.api_version, self.os_version = self.check_version(api_version or settings.ipf_version, base_url)
            self.base_url = (
                urljoin(base_url, f"api/{self.api_version}/")
                if not settings.ipf_dev
                else urljoin(base_url, f"{self.api_version}/")
            )
            token = token or settings.ipf_token
            username = username or settings.ipf_username
            password = password or settings.ipf_password

        if not token and not (username and password):
            logger.error("IP Fabric Token or Username/Password not provided.")
            raise RuntimeError("IP Fabric Token or Username/Password not provided.")

        self.auth = (
            HeaderApiKey(token) if token else PasswordCredentials(base_url, username, password, self.api_version)
        )
        # Get Current User, by doing that we are also ensuring the token is valid
        self.user = self.get_user()
        self.snapshots = self.get_snapshots()
        self.snapshot_id = snapshot_id
        logger.debug(f"Successfully connected to '{self.base_url.host}' IPF version '{self.os_version}' "
                     f"as user '{self.user.username}'")

    def get_user(self):
        """
        Gets current logged in user information.
        :return: User: User model of logged in user
        """
        resp = self.get("users/me")
        resp.raise_for_status()
        return User(**resp.json())

    def check_version(self, api_version, base_url):
        """
        Checks API Version and returns the version to use in the URL and the OS Version
        :param api_version: str: User defined API Version or None
        :param base_url: str: URL of IP Fabric
        :return: api_version, os_version
        """
        if api_version == "v1":
            raise RuntimeError("IP Fabric Version < 5.0 support has been dropped, please use ipfabric==4.4.3")
        api_version = api_version.lstrip('v').split(".") if api_version else \
            importlib_metadata.version("ipfabric").lstrip('v').split(".")

        resp = self.get(urljoin(base_url, "api/version"), headers={"Content-Type": "application/json"})
        resp.raise_for_status()
        os_api_version = resp.json()["apiVersion"].lstrip('v').split(".")
        if api_version[0:2] > os_api_version[0:2]:
            logger.warning(
                f"Specified API or SDK Version ({'.'.join(api_version)}) is greater then "
                f"OS API Version. Using OS Version:  ({'.'.join(os_api_version)})"
            )
            api_version = os_api_version
        elif os_api_version[0] > api_version[0]:
            raise RuntimeError(
                f"OS Major Version {os_api_version[0]} is greater then SDK Version "
                f"{api_version[0]}.  Please upgrade the Python SDK to the new major version."
            )

        return f"v{api_version[0]}.{api_version[1]}", resp.json()["releaseVersion"]

    def update(self):
        self.snapshots = self.get_snapshots()

    @property
    def loaded_snapshots(self):
        return {k: v for k, v in self.snapshots.items() if v.loaded}

    @property
    def unloaded_snapshots(self):
        return {k: v for k, v in self.snapshots.items() if not v.loaded}

    @property
    def snapshot_id(self):
        return self._snapshot_id

    @snapshot_id.setter
    def snapshot_id(self, snapshot_id):
        snapshot_id = snapshot_id or DEFAULT_ID
        if not self.loaded_snapshots:
            logger.warning("No Snapshots are currently loaded.  Please load a snapshot before querying any data.")
            self._snapshot_id = None
        elif snapshot_id not in self.snapshots:
            # Verify snapshot ID is valid
            logger.exception(f"Incorrect Snapshot ID: '{snapshot_id}'")
            raise ValueError(f"Incorrect Snapshot ID: '{snapshot_id}'")
        else:
            self._snapshot_id = self.snapshots[snapshot_id].snapshot_id

    def get_snapshots(self):
        """
        Gets all snapshots from IP Fabric and returns a dictionary of {ID: Snapshot_info}
        :return: dict[str, Snapshot]: Dictionary with ID as key and dictionary with info as the value
        """
        res = self.get("/snapshots")
        res.raise_for_status()

        snap_dict = OrderedDict()
        for s in res.json():
            snap = models.Snapshot(**s)
            snap_dict[snap.snapshot_id] = snap
            if snap.loaded:
                if "$lastLocked" not in snap_dict and snap.locked:
                    snap_dict["$lastLocked"] = snap
                if DEFAULT_ID not in snap_dict:
                    snap_dict[DEFAULT_ID] = snap
                    continue
                if "$prev" not in snap_dict:
                    snap_dict["$prev"] = snap
        return snap_dict



import logging
from pydantic import BaseModel
from typing import Any
from ipfabric import models

logger = logging.getLogger("python-ipfabric")


class Sdwan(BaseModel):
    client: Any

    @property
    def sites(self):
        return models.Table(client=self.client, endpoint="tables/sdwan/sites")

    @property
    def links(self):
        return models.Table(client=self.client, endpoint="tables/sdwan/links")



import logging
from pydantic import BaseModel
from typing import Any
from ipfabric import models

logger = logging.getLogger("python-ipfabric")


class Addressing(BaseModel):
    client: Any

    @property
    def arp_table(self):
        return models.Table(client=self.client, endpoint="tables/addressing/arp")

    @property
    def mac_table(self):
        return models.Table(client=self.client, endpoint="tables/addressing/mac")

    @property
    def managed_ip_ipv4(self):
        return models.Table(client=self.client, endpoint="tables/addressing/managed-devs")

    @property
    def managed_ip_ipv6(self):
        return models.Table(client=self.client, endpoint="tables/addressing/ipv6-managed-devs")

    @property
    def managed_duplicate_ip(self):
        return models.Table(client=self.client, endpoint="tables/addressing/duplicate-ip")

    @property
    def nat_rules(self):
        return models.Table(client=self.client, endpoint="tables/addressing/nat/rules")

    @property
    def nat_pools(self):
        return models.Table(client=self.client, endpoint="tables/addressing/nat/pools")

    @property
    def ipv6_neighbor_discovery(self):
        return models.Table(client=self.client, endpoint="tables/addressing/ipv6-neighbors")



import logging
from pydantic import BaseModel
from typing import Any
from ipfabric import models

logger = logging.getLogger("python-ipfabric")


class Neighbors(BaseModel):
    client: Any

    @property
    def neighbors_all(self):
        return models.Table(client=self.client, endpoint="tables/neighbors/all")

    @property
    def neighbors_unmanaged(self):
        return models.Table(client=self.client, endpoint="tables/neighbors/unmanaged")

    @property
    def neighbors_unidirectional(self):
        return models.Table(client=self.client, endpoint="tables/neighbors/unidirectional")

    @property
    def neighbors_endpoints(self):
        return models.Table(client=self.client, endpoint="tables/neighbors/endpoints")



import logging
from pydantic import BaseModel
from typing import Any
from ipfabric import models

logger = logging.getLogger("python-ipfabric")


class Mpls(BaseModel):
    client: Any

    @property
    def ldp_neighbors(self):
        return models.Table(client=self.client, endpoint="tables/mpls/ldp/neighbors")

    @property
    def ldp_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/mpls/ldp/interfaces")

    @property
    def rsvp_neighbors(self):
        return models.Table(client=self.client, endpoint="tables/mpls/rsvp/neighbors")

    @property
    def rsvp_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/mpls/rsvp/interfaces")

    @property
    def rsvp_forwarding(self):
        return models.Table(client=self.client, endpoint="tables/mpls/forwarding")

    @property
    def l3vpn_pe_routers(self):
        return models.Table(client=self.client, endpoint="tables/mpls/l3-vpn/pe-routers")

    @property
    def l3vpn_pe_vrfs(self):
        return models.Table(client=self.client, endpoint="tables/mpls/l3-vpn/pe-vrfs")

    @property
    def l3vpn_vrf_targets(self):
        return models.Table(client=self.client, endpoint="tables/mpls/l3-vpn/vrf-targets")

    @property
    def l3vpn_pe_routes(self):
        return models.Table(client=self.client, endpoint="tables/mpls/l3-vpn/pe-routes")

    @property
    def l2vpn_point_to_point_vpws(self):
        return models.Table(client=self.client, endpoint="tables/mpls/l2-vpn/point-to-point-vpws")

    @property
    def l2vpn_point_to_multipoint(self):
        return models.Table(client=self.client, endpoint="tables/mpls/l2-vpn/point-to-multipoint")

    @property
    def l2vpn_circuit_cross_connect(self):
        return models.Table(client=self.client, endpoint="tables/mpls/l2-vpn/curcit-cross-connect")

    @property
    def l2vpn_pseudowires(self):
        return models.Table(client=self.client, endpoint="tables/mpls/l2-vpn/pseudowires")



import logging
from pydantic import BaseModel
from typing import Any
from ipfabric import models

logger = logging.getLogger("python-ipfabric")


class Oam(BaseModel):
    client: Any

    @property
    def unidirectional_link_detection_neighbors(self):
        return models.Table(
            client=self.client, endpoint="tables/management/oam/unidirectional-link-detection/neighbors"
        )

    @property
    def unidirectional_link_detection_interfaces(self):
        return models.Table(
            client=self.client, endpoint="tables/management/oam/unidirectional-link-detection/interfaces"
        )



import logging
from datetime import datetime
from typing import Optional, Any, List

from pydantic import BaseModel, Field
from ipfabric import models

logger = logging.getLogger("python-ipfabric")


class Routing(BaseModel):
    client: Any

    @property
    def summary_protocols(self):
        return models.Table(client=self.client, endpoint="tables/networks/summary/protocols")

    @property
    def summary_protocols_bgp(self):
        return models.Table(client=self.client, endpoint="tables/networks/summary/protocols/bgp")

    @property
    def summary_protocols_eigrp(self):
        return models.Table(client=self.client, endpoint="tables/networks/summary/protocols/eigrp")

    @property
    def summary_protocols_isis(self):
        return models.Table(client=self.client, endpoint="tables/networks/summary/protocols/isis")

    @property
    def summary_protocols_ospf(self):
        return models.Table(client=self.client, endpoint="tables/networks/summary/protocols/ospf")

    @property
    def summary_protocols_ospfv3(self):
        return models.Table(client=self.client, endpoint="tables/networks/summary/protocols/ospfv3")

    @property
    def summary_protocols_rip(self):
        return models.Table(client=self.client, endpoint="tables/networks/summary/protocols/rip")

    @property
    def routes_ipv4(self):
        return models.Table(client=self.client, endpoint="tables/networks/routes")

    @property
    def routes_ipv6(self):
        return models.Table(client=self.client, endpoint="tables/networks/ipv6-routes")

    @property
    def route_stability(self):
        return models.Table(client=self.client, endpoint="tables/networks/route-stability")

    @property
    def ospf_neighbors(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/ospf/neighbors")

    @property
    def ospf_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/ospf/interfaces")

    @property
    def ospfv3_neighbors(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/ospf-v3/neighbors")

    @property
    def ospfv3_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/ospf-v3/interfaces")

    @property
    def bgp_neighbors(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/bgp/neighbors")

    @property
    def bgp_address_families(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/bgp/address-families")

    @property
    def eigrp_neighbors(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/eigrp/neighbors")

    @property
    def eigrp_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/eigrp/interfaces")

    @property
    def rip_neighbors(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/rip/neighbors")

    @property
    def rip_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/rip/interfaces")

    @property
    def isis_neighbors(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/is-is/neighbors")

    @property
    def isis_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/routing/protocols/is-is/interfaces")

    @property
    def path_lookup_checks(self):
        return models.Table(client=self.client, endpoint="tables/networks/path-lookup-checks")

    @property
    def vrf_summary(self):
        return models.Table(client=self.client, endpoint="tables/vrf/summary")

    @property
    def vrf_detail(self):
        return models.Table(client=self.client, endpoint="tables/vrf/detail")

    @property
    def vrf_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/vrf/interfaces")



import logging
from pydantic import BaseModel
from typing import Any
from ipfabric import models

logger = logging.getLogger("python-ipfabric")


class Wireless(BaseModel):
    client: Any

    @property
    def controllers(self):
        return models.Table(client=self.client, endpoint="tables/wireless/controllers")

    @property
    def access_points(self):
        return models.Table(client=self.client, endpoint="tables/wireless/access-points")

    @property
    def radios_detail(self):
        return models.Table(client=self.client, endpoint="tables/wireless/radio")

    @property
    def radios_ssid_summary(self):
        return models.Table(client=self.client, endpoint="tables/wireless/ssid-summary")

    @property
    def clients(self):
        return models.Table(client=self.client, endpoint="tables/wireless/clients")



import csv
import logging
from os.path import exists
from typing import Union, Any

from pydantic import Field
from pydantic.dataclasses import dataclass

logger = logging.getLogger("python-ipfabric")


@dataclass
class UpdateSiteNames:
    ipf: Any
    sites: Union[str, list] = Field(description="List of tuples [(oldName, newName)] or CSV filename to import.")
    dry_run: bool = False

    def __post_init__(self):
        if isinstance(self.sites, str) and exists(self.sites):
            with open(self.sites, "r") as f:
                self.sites = list(csv.reader(f, delimiter=","))
        elif not isinstance(self.sites, list) and not (isinstance(self.sites, str) and exists(self.sites)):
            raise SyntaxError("Sites is not a list of tuples or a filename.")

    def _patch_site(self, key, old, new):
        if self.dry_run:
            return True
        res = self.ipf.patch("sites/" + key, json=dict(name=new))
        if res.status_code == 200:
            logger.warning(f"Changed {old} to {new}")
            return True
        else:
            logger.error(f"Failed to change {old} to {new}")
            return False

    def update_sites(self):
        """
        Applies updates to IP Fabric
        :return: dict: {updated: [oldName, newName], errors: [oldName, newName]}
        """
        ipf_sites = self.ipf.inventory.sites.all(columns=["siteName", "siteKey", "siteUid"])
        names = {s["siteName"]: s for s in ipf_sites}
        uid = {s["siteUid"]: s for s in ipf_sites}
        updates = dict(updated=list(), errors=list())
        for site in self.sites:
            old, new = site[0].strip(), site[1].strip()
            changed = False
            if old in names and names[old]["siteName"] != new:
                changed = self._patch_site(names[old]["siteKey"], old, new)
            elif old in uid and uid[old]["siteName"] != site[1]:
                changed = self._patch_site(uid[old]["siteKey"], uid[old]["siteName"], new)
            elif (old in names and names[old]["siteName"] == new) or (old in uid and uid[old]["siteName"] == new):
                logger.info(f"{old} is already set to {new}")
            else:
                logger.error(f"Could not find {old}")
            if changed:
                updates["updated"].append((old, new))
            else:
                updates["errors"].append((old, new))
        return updates



import logging
from pydantic import BaseModel
from typing import Any
from ipfabric import models

logger = logging.getLogger("python-ipfabric")


class Qos(BaseModel):
    client: Any

    @property
    def policy_maps(self):
        return models.Table(client=self.client, endpoint="tables/qos/policy-maps")

    @property
    def shapping(self):
        return models.Table(client=self.client, endpoint="tables/qos/shapping")

    @property
    def queuing(self):
        return models.Table(client=self.client, endpoint="tables/qos/queuing")

    @property
    def policing(self):
        return models.Table(client=self.client, endpoint="tables/qos/policing")

    @property
    def priority_queuing(self):
        return models.Table(client=self.client, endpoint="tables/qos/priority-queuing")

    @property
    def marking(self):
        return models.Table(client=self.client, endpoint="tables/qos/marking")

    @property
    def random_drops(self):
        return models.Table(client=self.client, endpoint="tables/qos/random-drops")



import logging
from typing import Any

from pydantic.dataclasses import dataclass

logger = logging.getLogger("python-ipfabric")


@dataclass
class APIToken:
    client: Any

    @property
    def tokens(self):
        res = self.client.get("api-tokens")
        res.raise_for_status()
        return res.json()

    # def add_token(
    #     self,
    #     descr: str,
    #     scope: list,
    #     expires: Union[None, str] = None,
    #     token: str = None,
    # ):  # TODO: UPDATE FOR RBAC
    #     if token and len(token) < 8:
    #         raise SyntaxError("Token must be 8 characters.")
    #     elif not token:
    #         token = token_hex(16)
    #
    #     if not all(x in ["read", "write", "settings", "team"] for x in scope):
    #         raise SyntaxError("Only accepted scopes are ['read', 'write', 'settings', 'team']")
    #
    #     payload = dict(
    #         description=descr,
    #         expires=int(parser.parse(expires).timestamp() * 1000) if expires else None,
    #         scope=scope,
    #         token=token,
    #     )
    #     res = self.client.post("api-tokens", json=payload)
    #     return dict(token=token, response=res.json())

    def delete_token(self, token_id: str):
        res = self.client.delete("api-tokens/" + token_id)
        res.raise_for_status()
        return self.tokens



import logging
from pydantic import BaseModel
from typing import Any
from ipfabric import models

logger = logging.getLogger("python-ipfabric")


class LoadBalancing(BaseModel):
    client: Any

    @property
    def virtual_servers(self):
        return models.Table(client=self.client, endpoint="tables/load-balancing/virtual-servers")

    @property
    def virtual_servers_pools(self):
        return models.Table(client=self.client, endpoint="tables/load-balancing/virtual-servers/pools")

    @property
    def virtual_servers_pool_members(self):
        return models.Table(client=self.client, endpoint="tables/load-balancing/virtual-servers/pool-members")

    @property
    def virtual_servers_f5_partitions(self):
        return models.Table(client=self.client, endpoint="tables/load-balancing/virtual-servers/f5-partitions")



import logging
from typing import Any

from pydantic import BaseModel

from ipfabric import models

logger = logging.getLogger("python-ipfabric")


class Multicast(BaseModel):
    client: Any

    @property
    def pim_neighbors(self):
        return models.Table(client=self.client, endpoint="tables/multicast/pim/neighbors")

    @property
    def pim_neighbors(self):
        return models.Table(client=self.client, endpoint="tables/multicast/pim/neighbors")

    @property
    def mroute_overview(self):
        return models.Table(client=self.client, endpoint="tables/multicast/routes/overview")

    @property
    def mroute_table(self):
        return models.Table(client=self.client, endpoint="tables/multicast/routes/table")

    @property
    def mroute_oil_detail(self):
        return models.Table(client=self.client, endpoint="tables/multicast/routes/outgoing-interfaces")

    @property
    def mroute_counters(self):
        return models.Table(client=self.client, endpoint="tables/multicast/routes/counters")

    @property
    def mroute_first_hop_router(self):
        return models.Table(client=self.client, endpoint="tables/multicast/routes/first-hop-router")

    @property
    def mroute_sources(self):
        return models.Table(client=self.client, endpoint="tables/multicast/routes/sources")

    @property
    def igmp_groups(self):
        return models.Table(client=self.client, endpoint="tables/multicast/igmp/groups")

    @property
    def igmp_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/multicast/igmp/interfaces")

    @property
    def igmp_snooping_global_config(self):
        return models.Table(client=self.client, endpoint="tables/multicast/igmp/snooping/global")

    @property
    def igmp_snooping_groups(self):
        return models.Table(client=self.client, endpoint="tables/multicast/igmp/snooping/groups")

    @property
    def igmp_snooping_vlans(self):
        return models.Table(client=self.client, endpoint="tables/multicast/igmp/snooping/vlans")

    @property
    def mac_table(self):
        return models.Table(client=self.client, endpoint="tables/multicast/mac")

    @property
    def rp_overview(self):
        return models.Table(client=self.client, endpoint="tables/multicast/pim/rp/overview")

    @property
    def rp_bsr(self):
        return models.Table(client=self.client, endpoint="tables/multicast/pim/rp/bsr")

    @property
    def rp_mappings(self):
        return models.Table(client=self.client, endpoint="tables/multicast/pim/rp/mappings")

    @property
    def rp_mappings_groups(self):
        return models.Table(client=self.client, endpoint="tables/multicast/pim/rp/mappings-groups")



import logging
from typing import Any, Union, List

from ipfabric.intent_models import Group
from .intent_models import IntentCheck

logger = logging.getLogger("python-ipfabric")
COLOR_DICT = dict(green=0, blue=10, amber=20, red=30)


class Intent:
    def __init__(self, client):
        self.client: Any = client
        self.intent_checks: List[IntentCheck] = list()
        self.groups: List[Group] = list()
        self.snapshot_id: str = self.client.snapshot_id

    def get_intent_checks(self, snapshot_id: str = None):
        """
        Gets all intent checks and returns a list of them.  You can also:
            ipf.load_intent()  # Loads the intents to intent_checks
            print(len(ipf.intent.intent_checks))
        :param snapshot_id: str: Optional snapshot ID to get different data
        :return: list: List of intent checks
        """
        snapshot_id = self.client.snapshots[snapshot_id].snapshot_id if snapshot_id else self.client.snapshot_id
        res = self.client.get("reports", params=dict(snapshot=snapshot_id))
        res.raise_for_status()
        return [IntentCheck(**check) for check in res.json()]

    def load_intent(self, snapshot_id: str = None):
        """
        Loads intent checks into the class.
        :param snapshot_id: str: Uses a different Snapshot ID then client
        :return:
        """
        self.snapshot_id = snapshot_id or self.snapshot_id
        self.intent_checks = self.get_intent_checks(snapshot_id)
        self.groups = self.get_groups()

    def get_groups(self):
        res = self.client.get("reports/groups")
        res.raise_for_status()
        return [Group(**group) for group in res.json()]

    @property
    def custom(self):
        if not self.intent_checks:
            self.load_intent()
        return [c for c in self.intent_checks if c.custom]

    @property
    def builtin(self):
        if not self.intent_checks:
            self.load_intent()
        return [c for c in self.intent_checks if not c.custom]

    @property
    def intent_by_id(self):
        if not self.intent_checks:
            self.load_intent()
        return {c.intent_id: c for c in self.intent_checks}

    @property
    def intent_by_name(self):
        if not self.intent_checks:
            self.load_intent()
        return {c.name: c for c in self.intent_checks}

    @property
    def group_by_id(self):
        if not self.groups:
            self.load_intent()
        return {g.group_id: g for g in self.groups}

    @property
    def group_by_name(self):
        if not self.groups:
            self.load_intent()
        return {g.name: g for g in self.groups}

    def get_results(self, intent: IntentCheck, color: Union[str, int], snapshot_id: str = None):
        if isinstance(color, str):
            color = COLOR_DICT[color]
        snapshot_id = snapshot_id or self.snapshot_id
        return self._get_data(intent, snapshot_id, color)

    def get_all_results(self, intent: IntentCheck, snapshot_id: str = None):
        snapshot_id = snapshot_id or self.snapshot_id
        for color_str, color_int in COLOR_DICT.items():
            if getattr(intent.result.checks, color_str):
                setattr(intent.result_data, color_str, self._get_data(intent, snapshot_id, color_int))
        return intent

    def _get_data(self, intent: IntentCheck, snapshot_id: str, color: int):
        return self.client.fetch_all(
            intent.api_endpoint,
            snapshot_id=snapshot_id,
            reports=intent.web_endpoint,
            filters={intent.column: ["color", "eq", color]},
        )

    def compare_snapshot(self, snapshot_id: str, reverse: bool = False):
        """
        Compares all intents against another snapshot.
        Current is the snapshot loaded into the class
        Other is the snapshot specified in this method.  Use reverse=True to flip them.
        :param snapshot_id: str: Snapshot ID to compare against this will be the "other" key
        :param reverse: bool: Default False, setting to true will flip current and other.
        :return: list: List of dictionaries
        """
        new_intents = {i.name: i for i in self.get_intent_checks(snapshot_id)}
        comparison = list()
        for name, intent in new_intents.items():
            old = self.intent_by_name[name].result
            compare = intent.result.compare(old) if reverse else old.compare(intent.result)
            for desc, value in compare.items():
                n = desc if desc != "count" else "total"
                comparison.append({"name": name, "id": intent.intent_id, "check": n, **value})
        return comparison



from typing import Optional, List

from pydantic import BaseModel, Field, validator, AnyHttpUrl

AWS_REGIONS = [
    "us-east-2",
    "us-east-1",
    "us-west-1",
    "us-west-2",
    "af-south-1",
    "ap-east-1",
    "ap-southeast-3",
    "ap-south-1",
    "ap-northeast-3",
    "ap-northeast-2",
    "ap-southeast-1",
    "ap-southeast-2",
    "ap-northeast-1",
    "ca-central-1",
    "eu-central-1",
    "eu-west-1",
    "eu-west-2",
    "eu-south-1",
    "eu-west-3",
    "eu-north-1",
    "me-south-1",
    "sa-east-1",
    "us-gov-east-1",
    "us-gov-west-1",
]


class SystemProxy(BaseModel):
    respectSystemProxyConfiguration: bool = True


class RejectUnauthorized(SystemProxy, BaseModel):
    rejectUnauthorized: bool = True


class UserAuthBaseUrl(BaseModel):
    username: str
    password: str
    baseUrl: AnyHttpUrl
    isEnabled: bool = Field(default=True, const=True)


class AWS(SystemProxy, BaseModel):
    apiKey: str
    apiSecret: str
    region: str
    assumeRole: Optional[str] = Field(default=None)
    isEnabled: bool = Field(default=True, const=True)
    type: str = Field(default="aws-ec2", const=True)

    @validator("region")
    def check_region(cls, r):
        if r.lower() not in AWS_REGIONS:
            raise ValueError(f"{r} is not a valid AWS Region")
        return r.lower()


class Azure(SystemProxy, BaseModel):
    clientId: str
    clientSecret: str
    subscriptionId: str
    tenantId: str
    isEnabled: bool = Field(default=True, const=True)
    type: str = Field(default="azure", const=True)


class CheckPointApiKey(RejectUnauthorized, BaseModel):
    apiKey: str
    baseUrl: AnyHttpUrl
    domains: Optional[List[str]] = Field(default_factory=list)
    isEnabled: bool = Field(default=True, const=True)
    type: str = Field(default="checkpoint-mgmt-api", const=True)


class CheckPointUserAuth(RejectUnauthorized, UserAuthBaseUrl, BaseModel):
    domains: Optional[List[str]] = Field(default_factory=list)
    type: str = Field(default="checkpoint-mgmt-api", const=True)


class CiscoFMC(RejectUnauthorized, UserAuthBaseUrl, BaseModel):
    type: str = Field(default="ciscofmc", const=True)


class Merakiv1(RejectUnauthorized, BaseModel):
    apiKey: str
    baseUrl: AnyHttpUrl
    organizations: Optional[List[str]] = Field(default_factory=list)
    isEnabled: bool = Field(default=True, const=True)
    apiVer: str = Field(default="v0", const=True)
    type: str = Field(default="meraki-v0", const=True)


class NSXT(RejectUnauthorized, UserAuthBaseUrl, BaseModel):
    type: str = Field(default="nsxT", const=True)


class SilverPeak(RejectUnauthorized, UserAuthBaseUrl, BaseModel):
    type: str = Field(default="nsxT", const=True)


class Versa(RejectUnauthorized, UserAuthBaseUrl, BaseModel):
    type: str = Field(default="versa", const=True)


class Viptela(RejectUnauthorized, UserAuthBaseUrl, BaseModel):
    type: str = Field(default="viptela", const=True)



import logging
from typing import Any, Optional, List

from pydantic.dataclasses import dataclass

logger = logging.getLogger("python-ipfabric")


@dataclass
class Attributes:
    client: Any

    def all(
        self,
        columns: list = None,
        filters: Optional[dict] = None,
        sort: Optional[dict] = None,
    ):
        """
        Gets all data from corresponding endpoint
        :param columns: list: Optional columns to return, default is all
        :param filters: dict: Optional filters
        :param snapshot_id: str: Optional snapshot ID to override class
        :param reports: str: String of frontend URL where the reports are displayed
        :param sort: dict: Dictionary to apply sorting: {"order": "desc", "column": "lastChange"}
        :return: list: List of Dictionaries
        """
        return self.client.fetch_all(
            "tables/global-attributes", columns=columns, filters=filters, sort=sort, snapshot=False
        )

    def set_site_by_sn(self, serial_number, site_name):
        """
        Set a single site by serial number
        :param serial_number: str: IP Fabric Unique Serial Number
        :param site_name: str: Site name for device.
        :return:
        """
        resp = self.client.post("attributes/global", json=dict(name="siteName", sn=serial_number, value=site_name))
        resp.raise_for_status()
        return resp.json()

    def set_sites_by_sn(self, sites: List[dict]):
        """
        Sets a list of sites for devices based on serial numbers.
        :param sites: list: [{'sn': 'IPF SERIAL NUMBER', 'value': 'SITE NAME'}]
        :return:
        """
        [a.update(dict(name="siteName")) for a in sites]
        resp = self.client.put("attributes/global", json=dict(attributes=sites))
        resp.raise_for_status()
        return resp.json()

    def delete_attribute_by_sn(self, *serial_numbers):
        """
        Deletes attributes by Unique IP Fabric Serial Number(s)
        :param serial_numbers: str: Serial Numbers
        :return:
        """
        serial_numbers = [str(i) for i in serial_numbers]
        resp = self.client.request("DELETE", "attributes/global/", json=dict(attributes=dict(sn=serial_numbers)))
        resp.raise_for_status()
        return True

    def delete_attribute_by_id(self, *attribute_ids):
        """
        Deletes attributes by Attribute ID(s)
        :param attribute_ids: str: Attribute IDs
        :return:
        """
        attribute_ids = [str(i) for i in attribute_ids]
        resp = self.client.request("DELETE", "attributes/global/", json=dict(attributes=dict(id=attribute_ids)))
        resp.raise_for_status()
        return True



import logging
from pydantic import BaseModel
from typing import Any
from ipfabric import models

logger = logging.getLogger("python-ipfabric")


class Stp(BaseModel):
    client: Any

    @property
    def stability(self):
        return models.Table(client=self.client, endpoint="tables/spanning-tree/topology")

    @property
    def bridges(self):
        return models.Table(client=self.client, endpoint="tables/spanning-tree/bridges")

    @property
    def instances(self):
        return models.Table(client=self.client, endpoint="tables/spanning-tree/instances")

    @property
    def vlans(self):
        return models.Table(client=self.client, endpoint="tables/spanning-tree/vlans")

    @property
    def ports(self):
        return models.Table(client=self.client, endpoint="tables/spanning-tree/ports")

    @property
    def neighbors(self):
        return models.Table(client=self.client, endpoint="tables/spanning-tree/neighbors")

    @property
    def guards(self):
        return models.Table(client=self.client, endpoint="tables/spanning-tree/guards")

    @property
    def inconsistencies(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/inconsistencies/summary")

    @property
    def inconsistencies_details(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/inconsistencies/details")

    @property
    def inconsistencies_ports_vlan_mismatch(self):
        return models.Table(
            client=self.client, endpoint="tables/spanning-tree/inconsistencies/neighbor-ports-vlan-mismatch"
        )

    @property
    def inconsistencies_ports_multiple_neighbors(self):
        return models.Table(
            client=self.client, endpoint="tables/spanning-tree/inconsistencies/ports-multiple-neighbors"
        )

    @property
    def inconsistencies_stp_cdp_ports_mismatch(self):
        return models.Table(client=self.client, endpoint="tables/spanning-tree/inconsistencies/stp-cdp-ports-mismatch")

    @property
    def inconsistencies_multiple_stp(self):
        return models.Table(client=self.client, endpoint="tables/spanning-tree/inconsistencies/multiple-stp")



import logging
from pydantic import BaseModel
from typing import Any
from ipfabric import models

logger = logging.getLogger("python-ipfabric")


class Platforms(BaseModel):
    client: Any

    @property
    def environment_power_supplies(self):
        return models.Table(client=self.client, endpoint="tables/inventory/power-supplies")

    @property
    def environment_power_supplies_fans(self):
        return models.Table(client=self.client, endpoint="tables/inventory/power-supplies-fans")

    @property
    def environment_fans(self):
        return models.Table(client=self.client, endpoint="tables/inventory/fans")

    @property
    def environment_modules(self):
        return models.Table(client=self.client, endpoint="tables/inventory/modules")

    @property
    def juniper_cluster(self):
        return models.Table(client=self.client, endpoint="tables/platforms/cluster/srx")

    @property
    def cisco_fex_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/platforms/fex/interfaces")

    @property
    def cisco_fex_modules(self):
        return models.Table(client=self.client, endpoint="tables/platforms/fex/modules")

    @property
    def cisco_vdc_devices(self):
        return models.Table(client=self.client, endpoint="tables/platforms/vdc/devices")

    @property
    def platform_cisco_vss(self):
        return models.Table(client=self.client, endpoint="tables/platforms/vss/overview")

    @property
    def cisco_vss_chassis(self):
        return models.Table(client=self.client, endpoint="tables/platforms/vss/chassis")

    @property
    def cisco_vss_vsl(self):
        return models.Table(client=self.client, endpoint="tables/platforms/vss/vsl")

    @property
    def poe_devices(self):
        return models.Table(client=self.client, endpoint="tables/platforms/poe/devices")

    @property
    def poe_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/platforms/poe/interfaces")

    @property
    def poe_modules(self):
        return models.Table(client=self.client, endpoint="tables/platforms/poe/modules")

    @property
    def stacks(self):
        return models.Table(client=self.client, endpoint="tables/platforms/stacks")

    @property
    def stacks_members(self):
        return models.Table(client=self.client, endpoint="tables/platforms/stack/members")

    @property
    def stacks_stack_ports(self):
        return models.Table(client=self.client, endpoint="tables/platforms/stack/connections")



from typing import Optional

from pydantic import BaseModel

from ipfabric.tools.nist import NIST, CVEs


class Version(BaseModel):
    vendor: str
    family: Optional[str]
    version: str
    cves: CVEs
    hostname: Optional[str]
    site: Optional[str]


class Vulnerabilities:
    def __init__(self, ipf, timeout: int = 30, cve_limit: int = 20):
        self.ipf = ipf
        self.nist = NIST(timeout=timeout, cve_limit=cve_limit)

    def __del__(self):
        try:
            self.nist.close()
        except AttributeError:
            pass

    def _check_versions(self, versions):
        cves = list()
        for v in versions:
            cve = self.nist.check_cve(v["vendor"], v["family"], v["version"])
            version = Version(vendor=v["vendor"], family=v["family"], version=v["version"], cves=cve)
            if "hostname" in v:
                version.hostname = v["hostname"]
                version.site = v["siteName"]
            cves.append(version)
        return cves

    def check_versions(self, vendor=None):
        filters = {"vendor": ["like", vendor]} if vendor else None
        versions = self.ipf.fetch_all(
            "tables/management/osver-consistency",
            columns=["vendor", "family", "version"],
            filters=filters,
        )
        return self._check_versions(versions)

    def check_device(self, device):
        devices = self.ipf.inventory.devices.all(
            columns=["hostname", "siteName", "vendor", "family", "version"],
            filters={"hostname": ["like", device]},
        )
        return self._check_versions(devices)

    def check_site(self, site):
        sites = self.ipf.inventory.devices.all(
            columns=["hostname", "siteName", "vendor", "family", "version"],
            filters={"siteName": ["like", site]},
        )
        return self._check_versions(sites)



import logging
from typing import Any

from pydantic import BaseModel

from ipfabric import models

logger = logging.getLogger("python-ipfabric")


class ManagedNetworks(BaseModel):
    client: Any

    @property
    def networks(self):
        return models.Table(client=self.client, endpoint="tables/networks")

    @property
    def gateway_redundancy(self):
        return models.Table(client=self.client, endpoint="tables/networks/gateway-redundancy")



import re
from json import loads
from typing import Optional, Union
from urllib.parse import urlparse

from ipfabric.api import IPFabricAPI
from ipfabric.intent import Intent
from ipfabric.models import Technology, Inventory

DEFAULT_ID = "$last"


def check_format(func):
    """
    Checks to make sure api/v1/ is not in the URL and converts filters from json str to dict
    """

    def wrapper(self, url, *args, **kwargs):
        if "filters" in kwargs and isinstance(kwargs["filters"], str):
            kwargs["filters"] = loads(kwargs["filters"])
        path = urlparse(url or kwargs["url"]).path
        r = re.search(r"(api/)?v\d(\.\d)?/", path)
        url = path[r.end() :] if r else path
        url = url[1:] if url[0] == "/" else url
        return func(self, url, *args, **kwargs)

    return wrapper


class IPFClient(IPFabricAPI):
    def __init__(
        self,
        base_url: Optional[str] = None,
        api_version: Optional[str] = None,
        token: Optional[str] = None,
        snapshot_id: str = DEFAULT_ID,
        username: Optional[str] = None,
        password: Optional[str] = None,
        **kwargs,
    ):
        """
        Initializes the IP Fabric Client
        :param base_url: str: IP Fabric instance provided in 'base_url' parameter, or the 'IPF_URL' environment variable
        :param token: str: API token or 'IPF_TOKEN' environment variable
        :param snapshot_id: str: IP Fabric snapshot ID to use by default for database actions - defaults to '$last'
        :param kwargs: dict: Keyword args to pass to httpx
        """
        super().__init__(base_url, api_version, token, snapshot_id, username, password, **kwargs)
        self.inventory = Inventory(client=self)
        self.intent = Intent(client=self)
        self.technology = Technology(client=self)

    @check_format
    def fetch(
        self,
        url,
        columns: Optional[list] = None,
        filters: Optional[Union[dict, str]] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        snapshot_id: Optional[str] = None,
        reports: Optional[str] = None,
        sort: Optional[dict] = None,
        snapshot: bool = True,
    ):
        """
        Gets data from IP Fabric for specified endpoint
        :param url: str: Example tables/vlan/device-summary
        :param columns: list: Optional list of columns to return, None will return all
        :param filters: dict: Optional dictionary of filters
        :param limit: int: Default to 1,000 rows
        :param start: int: Starts at 0
        :param snapshot_id: str: Optional snapshot_id to override default
        :param reports: str: String of frontend URL where the reports are displayed
        :param sort: dict: Dictionary to apply sorting: {"order": "desc", "column": "lastChange"}
        :param snapshot: bool: Set to False for some tables like management endpoints.
        :return: list: List of Dictionary objects.
        """

        payload = dict(
            columns=columns or self._get_columns(url),
            pagination=dict(start=start, limit=limit),
        )
        if snapshot:
            payload["snapshot"] = snapshot_id or self.snapshot_id
        if filters:
            payload["filters"] = filters
        if reports:
            payload["reports"] = reports
        if sort:
            payload["sort"] = sort

        res = self.post(url, json=payload)
        res.raise_for_status()
        return res.json()["data"]

    @check_format
    def fetch_all(
        self,
        url: str,
        columns: Optional[list] = None,
        filters: Optional[Union[dict, str]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[str] = None,
        sort: Optional[dict] = None,
        snapshot: bool = True,
    ):
        """
        Gets all data from IP Fabric for specified endpoint
        :param url: str: Example tables/vlan/device-summary
        :param columns: list: Optional list of columns to return, None will return all
        :param filters: dict: Optional dictionary of filters
        :param snapshot_id: str: Optional snapshot_id to override default
        :param reports: str: String of frontend URL where the reports are displayed
        :param sort: dict: Dictionary to apply sorting: {"order": "desc", "column": "lastChange"}
        :param snapshot: bool: Set to False for some tables like management endpoints.
        :return: list: List of Dictionary objects.
        """

        payload = dict(columns=columns or self._get_columns(url))
        if snapshot:
            payload["snapshot"] = snapshot_id or self.snapshot_id
        if filters:
            payload["filters"] = filters
        if reports:
            payload["reports"] = reports
        if sort:
            payload["sort"] = sort

        return self._ipf_pager(url, payload)

    @check_format
    def query(self, url: str, payload: Union[str, dict], all: bool = True):
        """
        Submits a query, does no formatting on the parameters.  Use for copy/pasting from the webpage.
        :param url: str: Example: https://demo1.ipfabric.io/api/v1/tables/vlan/device-summary
        :param payload: Union[str, dict]: Dictionary to submit in POST or can be JSON string (i.e. read from file).
        :param all: bool: Default use pager to get all results and ignore pagination information in the payload
        :return: list: List of Dictionary objects.
        """
        if isinstance(payload, str):
            payload = loads(payload)
        if all:
            return self._ipf_pager(url, payload)
        else:
            res = self.post(url, json=payload)
            res.raise_for_status()
            return res.json()["data"]

    def _get_columns(self, url: str):
        """
        Submits malformed payload and extracts column names from it
        :param url: str: API url to post
        :return: list: List of column names
        """
        r = self.post(url, json=dict(snapshot=self.snapshot_id, columns=["*"]))
        if r.status_code == 422:
            msg = r.json()["errors"][0]["message"]
            return [x.strip() for x in re.match(r"\".*\".*\[(.*)]$", msg).group(1).split(",")]
        else:
            r.raise_for_status()

    def _ipf_pager(
        self,
        url: str,
        payload: dict,
        data: Optional[Union[list, None]] = None,
        limit: int = 1000,
        start: int = 0,
    ):
        """
        Loops through and collects all the data from the tables
        :param url: str: Full URL to post to
        :param payload: dict: Data to submit to IP Fabric
        :param data: list: List of data to append subsequent calls
        :param start: int: Where to start for the data
        :return: list: List of dictionaries
        """
        data = data or list()

        payload["pagination"] = dict(limit=limit, start=start)
        r = self.post(url, json=payload)
        r.raise_for_status()
        r_data = r.json()["data"]
        data.extend(r_data)
        if limit == len(r_data):
            self._ipf_pager(url, payload, data, limit=limit, start=start + limit)
        return data

    def get_count(self, url: str, filters: Optional[Union[dict, str]] = None, snapshot_id: Optional[str] = None):
        payload = dict(columns=["id"], pagination=dict(limit=1, start=0), snapshot=snapshot_id or self.snapshot_id)
        if filters:
            payload["filters"] = filters
        res = self.post(url, json=payload)
        res.raise_for_status()
        return res.json()["_meta"]["count"]



import logging
from pydantic import BaseModel
from typing import Any
from ipfabric import models

logger = logging.getLogger("python-ipfabric")


class Fhrp(BaseModel):
    client: Any

    @property
    def group_state(self):
        return models.Table(client=self.client, endpoint="tables/fhrp/group-state")

    @property
    def group_members(self):
        return models.Table(client=self.client, endpoint="tables/fhrp/group-members")

    @property
    def stproot_alignment(self):
        return models.Table(client=self.client, endpoint="tables/fhrp/stproot-alignment")

    @property
    def balancing(self):
        return models.Table(client=self.client, endpoint="tables/fhrp/balancing")

    @property
    def glbp_forwarders(self):
        return models.Table(client=self.client, endpoint="tables/fhrp/glbp-forwarders")

    @property
    def virtual_gateways(self):
        return models.Table(client=self.client, endpoint="tables/fhrp/virtual-gateways")



from .addressing import Addressing
from .cloud import Cloud
from .dhcp import Dhcp
from .fhrp import Fhrp
from .interfaces import Interfaces
from .ip_telephony import IpTelephony
from .load_balancing import LoadBalancing
from .managed_networks import ManagedNetworks
from .management import Management
from .mpls import Mpls
from .multicast import Multicast
from .neighbors import Neighbors
from .oam import Oam
from .platforms import Platforms
from .port_channels import PortChannels
from .qos import Qos
from .routing import Routing
from .sdn import Sdn
from .sdwan import Sdwan
from .security import Security
from .stp import Stp
from .vlans import Vlans
from .wireless import Wireless

__all__ = [
    "Addressing",
    "Dhcp",
    "Fhrp",
    "Interfaces",
    "Management",
    "ManagedNetworks",
    "Mpls",
    "Multicast",
    "Neighbors",
    "Platforms",
    "PortChannels",
    "Routing",
    "Stp",
    "Vlans",
    "Cloud",
    "IpTelephony",
    "LoadBalancing",
    "Oam",
    "Qos",
    "Sdn",
    "Sdwan",
    "Security",
    "Wireless",
]



import logging
from pydantic import BaseModel
from typing import Any
from ipfabric import models

logger = logging.getLogger("python-ipfabric")


class Cloud(BaseModel):
    client: Any

    @property
    def virtual_machines(self):
        return models.Table(client=self.client, endpoint="tables/cloud/virtual-machines")

    @property
    def virtual_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/cloud/virtual-machines-interfaces")



import logging
from pydantic import BaseModel
from typing import Any
from ipfabric import models

logger = logging.getLogger("python-ipfabric")


class Dhcp(BaseModel):
    client: Any

    @property
    def relay_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/dhcp/relay/interfaces")

    @property
    def relay_interfaces_stats_received(self):
        return models.Table(client=self.client, endpoint="tables/dhcp/relay/interfaces-stats/received")

    @property
    def relay_interfaces_stats_relayed(self):
        return models.Table(client=self.client, endpoint="tables/dhcp/relay/interfaces-stats/relayed")

    @property
    def relay_interfaces_stats_sent(self):
        return models.Table(client=self.client, endpoint="tables/dhcp/relay/interfaces-stats/sent")

    @property
    def relay_global_stats_summary(self):
        return models.Table(client=self.client, endpoint="tables/dhcp/relay/global-stats/summary")

    @property
    def relay_global_stats_received(self):
        return models.Table(client=self.client, endpoint="tables/dhcp/relay/global-stats/received")

    @property
    def relay_global_stats_relayed(self):
        return models.Table(client=self.client, endpoint="tables/dhcp/relay/global-stats/relayed")

    @property
    def relay_global_stats_sent(self):
        return models.Table(client=self.client, endpoint="tables/dhcp/relay/global-stats/sent")

    @property
    def server_summary(self):
        return models.Table(client=self.client, endpoint="tables/dhcp/server/summary")

    @property
    def server_pools(self):
        return models.Table(client=self.client, endpoint="tables/dhcp/server/pools")

    @property
    def server_leases(self):
        return models.Table(client=self.client, endpoint="tables/dhcp/server/leases")

    @property
    def server_excluded_ranges(self):
        return models.Table(client=self.client, endpoint="tables/dhcp/server/excluded-ranges")

    @property
    def server_excluded_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/dhcp/server/interfaces")



import logging
from pydantic import BaseModel
from typing import Any
from ipfabric import models

logger = logging.getLogger("python-ipfabric")


class Security(BaseModel):
    client: Any

    @property
    def acl(self):
        return models.Table(client=self.client, endpoint="tables/security/acl")

    @property
    def acl_interface(self):
        return models.Table(client=self.client, endpoint="tables/security/acl/interfaces")

    @property
    def acl_global_policies(self):
        return models.Table(client=self.client, endpoint="tables/security/acl/global-policies")

    @property
    def dmvpn(self):
        return models.Table(client=self.client, endpoint="tables/security/dmvpn")

    @property
    def dhcp_snooping(self):
        return models.Table(client=self.client, endpoint="tables/security/dhcp/snooping")

    @property
    def dhcp_snooping_bindings(self):
        return models.Table(client=self.client, endpoint="tables/security/dhcp/bindings")

    @property
    def ipsec_tunnels(self):
        return models.Table(client=self.client, endpoint="tables/security/ipsec/tunnels")

    @property
    def ipsec_gateways(self):
        return models.Table(client=self.client, endpoint="tables/security/ipsec/gateways")

    @property
    def secure_ports_devices(self):
        return models.Table(client=self.client, endpoint="tables/security/secure-ports/devices")

    @property
    def secure_ports_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/security/secure-ports/interfaces")

    @property
    def secure_ports_users(self):
        return models.Table(client=self.client, endpoint="tables/security/secure-ports/users")

    @property
    def zone_firewall_policies(self):
        return models.Table(client=self.client, endpoint="tables/security/zone-firewall/policies")

    @property
    def zone_firewall_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/security/zone-firewall/interfaces")



from typing import List

from httpx import Client, ReadTimeout, HTTPStatusError
from pydantic import BaseModel, Field


class CVE(BaseModel):
    cve_id: str
    description: str
    url: str

    def __repr__(self):
        return self.cve_id

    def __hash__(self):
        return hash(self.cve_id)


class CVEs(BaseModel):
    total_results: int
    cves: List[CVE]
    error: str = Field(default=None)


class NIST(Client):
    def __init__(self, timeout, cve_limit):
        super().__init__(base_url="https://services.nvd.nist.gov/rest/json/cves/1.0", timeout=timeout)
        self.cve_limit = cve_limit

    @property
    def params(self):
        return {
            "cpeMatchString": "cpe:2.3:*:",
            "startIndex": 0,
            "resultsPerPage": self.cve_limit,
        }

    def check_cve(self, vendor: str, family: str, version: str):
        """

        :param vendor: str: Vendor for the device to be checked
        :param family: str: Family of the device to be checked
        :param version: str: Software version of the device to be checked
        :return:
        """
        params = self.params
        if vendor in ["azure", "aws"]:
            return CVEs(total_results=0, cves=[], error="Unsupported")
        elif vendor == "juniper":
            version = version[: version.rfind("R") + 2].replace("R", ":r")
            params["cpeMatchString"] += vendor + ":" + family + ":" + version
        elif vendor == "paloalto":
            params["cpeMatchString"] += "palo_alto" + ":" + family + ":" + version
        elif vendor == "extreme":
            if "xos" in family:
                family = "extremexos"
            params["cpeMatchString"] += "extremenetworks" + ":" + family + ":" + version
        elif "aruba" in family:
            params["cpeMatchString"] += "arubanetworks:arubaos" + ":" + version
        elif vendor == "f5" and family == "big-ip":
            params["cpeMatchString"] += vendor + ":" + "big-ip_access_policy_manager" + ":" + version
        elif vendor == "cisco":
            if family == "meraki":
                return CVEs(total_results=0, cves=[], error="Unsupported")
            elif family == "wlc-air":
                family = "wireless_lan_controller_software"
            elif family == "ftd":
                family = "firepower"
                version = (version.replace("(Bu", ".Bu")).replace(")", ".").split(" .", 1)[0]
            elif family != "nx-os":
                family = family.replace("-", "_")
            version = (version.replace("(", ".")).replace(")", ".")
            params["cpeMatchString"] += vendor + ":" + family + ":" + version
        elif vendor == "fortinet" and family == "fortigate":
            params["cpeMatchString"] += "fortinet:fortios:" + version.replace(",", ".")
        elif vendor == "checkpoint" and family == "gaia":
            params["cpeMatchString"] += vendor + ":" + "gaia_os" + version.replace("R", ":r")
        elif vendor == "arista":
            params["cpeMatchString"] += vendor + ":" + family + ":" + version.lower()
        else:
            return CVEs(total_results=0, cves=[], error="Unsupported")

        try:
            res = self.get("", params=params)
            res.raise_for_status()
            data = res.json()

            cves = CVEs(
                total_results=data["totalResults"],
                cves=[
                    CVE(
                        cve_id=i["cve"]["CVE_data_meta"]["ID"],
                        description=i["cve"]["description"]["description_data"][0]["value"],
                        url=i["cve"]["references"]["reference_data"][0]["url"],
                    )
                    for i in data["result"]["CVE_Items"]
                ],
            )
            return cves
        except ReadTimeout:
            return CVEs(total_results=0, cves=[], error="Timeout")
        except HTTPStatusError:
            return CVEs(total_results=0, cves=[], error="HTTP Error")



import logging
from pydantic import BaseModel
from typing import Any
from ipfabric import models

logger = logging.getLogger("python-ipfabric")


class Vlans(BaseModel):
    client: Any

    @property
    def device_summary(self):
        return models.Table(client=self.client, endpoint="tables/vlan/device-summary")

    @property
    def device_detail(self):
        return models.Table(client=self.client, endpoint="tables/vlan/device")

    @property
    def network_summary(self):
        return models.Table(client=self.client, endpoint="tables/vlan/network-summary")

    @property
    def site_summary(self):
        return models.Table(client=self.client, endpoint="tables/vlan/site-summary")

    @property
    def l3_gateways(self):
        return models.Table(client=self.client, endpoint="tables/vlan/l3-gateways")



import logging
from datetime import datetime
from typing import Optional, Any, List

from pydantic import BaseModel, Field

from ipfabric.technology import *

logger = logging.getLogger("python-ipfabric")


class Site(BaseModel):
    _sitename: str = Field(alias="siteName")
    _name: str = Field(alias="name")
    uid: str
    site_id: Optional[str] = Field(None, alias="id")

    @property
    def site_name(self):
        return self._name or self._sitename


class Error(BaseModel):
    error_type: str = Field(alias="errorType")
    count: int


class Snapshot(BaseModel):
    snapshot_id: str = Field(alias="id")
    name: Optional[str]
    note: Optional[str]
    count: int = Field(alias="totalDevCount")
    licensed_count: int = Field(alias="licensedDevCount")
    status: str
    state: str
    locked: bool
    start: datetime = Field(alias="tsStart")
    end: Optional[datetime] = Field(alias="tsEnd")
    version: Optional[str] = None
    sites: List[Site]
    errors: Optional[List[Error]]

    @property
    def loaded(self):
        return self.state == "loaded"

    def unload(self, ipf):
        """
        Load Snapshot
        :param ipf: IPFClient
        :return: True
        """
        if self.loaded:
            res = ipf.post(
                "snapshots/unload", json=[dict(jobDetail=int(datetime.now().timestamp() * 1000), id=self.snapshot_id)]
            )
            res.raise_for_status()
        else:
            logger.warning(f"Snapshot {self.snapshot_id} is already unloaded.")
        return True

    def load(self, ipf):
        """
        Load Snapshot
        :param ipf: IPFClient
        :return: True
        """
        if not self.loaded:
            res = ipf.post(
                "snapshots/load", json=[dict(jobDetail=int(datetime.now().timestamp() * 1000), id=self.snapshot_id)]
            )
            res.raise_for_status()
        else:
            logger.warning(f"Snapshot {self.snapshot_id} is already loaded.")
        return True

    def attributes(self, ipf):
        """
        Load Snapshot
        :param ipf: IPFClient
        :return: True
        """
        return ipf.fetch_all("tables/snapshot-attributes", snapshot_id=self.snapshot_id)


class Table(BaseModel):
    endpoint: str
    client: Any

    @property
    def name(self):
        return self.endpoint.split("/")[-1]

    def fetch(
        self,
        columns: list = None,
        filters: Optional[dict] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[str] = None,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
    ):
        """
        Gets all data from corresponding endpoint
        :param columns: list: Optional columns to return, default is all
        :param filters: dict: Optional filters
        :param snapshot_id: str: Optional snapshot ID to override class
        :param reports: str: String of frontend URL where the reports are displayed
        :param sort: dict: Dictionary to apply sorting: {"order": "desc", "column": "lastChange"}
        :param limit: int: Default to 1,000 rows
        :param start: int: Starts at 0
        :return: list: List of Dictionaries
        """
        return self.client.fetch(
            self.endpoint,
            columns=columns,
            filters=filters,
            snapshot_id=snapshot_id,
            reports=reports,
            sort=sort,
            limit=limit,
            start=start,
        )

    def all(
        self,
        columns: list = None,
        filters: Optional[dict] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[str] = None,
        sort: Optional[dict] = None,
    ):
        """
        Gets all data from corresponding endpoint
        :param columns: list: Optional columns to return, default is all
        :param filters: dict: Optional filters
        :param snapshot_id: str: Optional snapshot ID to override class
        :param reports: str: String of frontend URL where the reports are displayed
        :param sort: dict: Dictionary to apply sorting: {"order": "desc", "column": "lastChange"}
        :return: list: List of Dictionaries
        """
        return self.client.fetch_all(
            self.endpoint,
            columns=columns,
            filters=filters,
            snapshot_id=snapshot_id,
            reports=reports,
            sort=sort,
        )

    def count(self, filters: Optional[dict] = None, snapshot_id: Optional[str] = None):
        """
        Gets count of table
        :param filters: dict: Optional filters
        :param snapshot_id: str: Optional snapshot ID to override class
        :return: int: Count
        """
        return self.client.get_count(
            self.endpoint,
            filters=filters,
            snapshot_id=snapshot_id,
        )


class Inventory(BaseModel):
    client: Any

    @property
    def sites(self):
        return Table(client=self.client, endpoint="/tables/inventory/sites")

    @property
    def vendors(self):
        return Table(client=self.client, endpoint="/tables/inventory/summary/vendors")

    @property
    def devices(self):
        return Table(client=self.client, endpoint="/tables/inventory/devices")

    @property
    def models(self):
        return Table(client=self.client, endpoint="/tables/inventory/summary/models")

    @property
    def platforms(self):
        return Table(client=self.client, endpoint="/tables/inventory/summary/platforms")

    @property
    def pn(self):
        return Table(client=self.client, endpoint="/tables/inventory/pn")

    @property
    def families(self):
        return Table(client=self.client, endpoint="/tables/inventory/summary/families")

    @property
    def interfaces(self):
        return Table(client=self.client, endpoint="/tables/inventory/interfaces")

    @property
    def hosts(self):
        return Table(client=self.client, endpoint="tables/addressing/hosts")

    @property
    def phones(self):
        return Table(client=self.client, endpoint="tables/inventory/phones")

    @property
    def fans(self):
        return Table(client=self.client, endpoint="tables/inventory/fans")

    @property
    def modules(self):
        return Table(client=self.client, endpoint="tables/inventory/modules")

    @property
    def powerSupplies(self):
        logger.warning(
            """Use of client.inventory.PowerSupplies will be deprecated in a future release, please 
                        use client.technology.platform_environment_power_supplies"""
        )
        return Table(client=self.client, endpoint="tables/inventory/power-supplies")

    @property
    def powerSuppliesFans(self):
        logger.warning(
            """Use of client.inventory.PowerSuppliesFans will be deprecated in a future release, please 
                        use client.technology.platform_environment_power_supplies_fans"""
        )
        return Table(client=self.client, endpoint="tables/inventory/power-supplies-fans")


class Technology(BaseModel):
    client: Any

    @property
    def platforms(self):
        return Platforms(client=self.client)

    @property
    def interfaces(self):
        return Interfaces(client=self.client)

    @property
    def neighbors(self):
        return Neighbors(client=self.client)

    @property
    def dhcp(self):
        return Dhcp(client=self.client)

    @property
    def port_channels(self):
        return PortChannels(client=self.client)

    @property
    def vlans(self):
        return Vlans(client=self.client)

    @property
    def stp(self):
        return Stp(client=self.client)

    @property
    def addressing(self):
        return Addressing(client=self.client)

    @property
    def fhrp(self):
        return Fhrp(client=self.client)

    @property
    def managed_networks(self):
        return ManagedNetworks(client=self.client)

    @property
    def mpls(self):
        return Mpls(client=self.client)

    @property
    def multicast(self):
        return Multicast(client=self.client)

    @property
    def cloud(self):
        return Cloud(client=self.client)

    @property
    def management(self):
        return Management(client=self.client)

    @property
    def ip_telephony(self):
        return IpTelephony(client=self.client)

    @property
    def load_balancing(self):
        return LoadBalancing(client=self.client)

    @property
    def oam(self):
        return Oam(client=self.client)

    @property
    def qos(self):
        return Qos(client=self.client)

    @property
    def routing(self):
        return Routing(client=self.client)

    @property
    def sdn(self):
        return Sdn(client=self.client)

    @property
    def sdwan(self):
        return Sdwan(client=self.client)

    @property
    def security(self):
        return Security(client=self.client)

    @property
    def wireless(self):
        return Wireless(client=self.client)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ipfabric',
 'ipfabric.settings',
 'ipfabric.technology',
 'ipfabric.tools',
 'ipfabric.tools.factory_defaults',
 'ipfabric.tools.factory_defaults.v4',
 'ipfabric.tools.factory_defaults.v4.4',
 'ipfabric.tools.factory_defaults.v5',
 'ipfabric.tools.factory_defaults.v5.0',
 'ipfabric.tools.factory_defaults.v6',
 'ipfabric.tools.factory_defaults.v6.0']

package_data = \
{'': ['*']}

install_requires = \
['httpx>=0.23.0,<0.24.0',
 'ipfabric-httpx-auth>=5.0.0,<6.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'python-dotenv>=0.21,<0.22',
 'pytz>=2022.1,<2023.0']

extras_require = \
{'examples': ['pandas==1.3.0',
              'openpyxl>=3.0.9,<4.0.0',
              'tabulate>=0.8.9,<0.9.0',
              'python-json-logger>=2.0.4,<3.0.0']}

setup_kwargs = {
    'name': 'ipfabric',
    'version': '5.0.11',
    'description': 'Python package for interacting with IP Fabric',
    'long_description': '# IPFabric\n\nIPFabric is a Python module for connecting to and communicating against an IP Fabric instance.\n\n[![Requirements Status](https://requires.io/github/community-fabric/python-ipfabric/requirements.svg?branch=develop)](https://requires.io/github/community-fabric/python-ipfabric/requirements/?branch=develop)\n\n## About\n\nFounded in 2015, [IP Fabric](https://ipfabric.io/) develops network infrastructure visibility and analytics solution to\nhelp enterprise network and security teams with network assurance and automation across multi-domain heterogeneous\nenvironments. From in-depth discovery, through graph visualization, to packet walks and complete network history, IP\nFabric enables to confidently replace manual tasks necessary to handle growing network complexity driven by relentless\ndigital transformation. \n\n## Versioning\nStarting with IP Fabric version 5.0.x the python-ipfabric and python-ipfabric-diagrams will need to\nmatch your IP Fabric version.  The API\'s are changing and instead of `api/v1` they will now be `api/v5.0`.\n\nVersion 5.1 will have backwards compatability with version 5.0 however 6.0 will not support any 5.x versions.\nBy ensuring that your ipfabric SDK\'s match your IP Fabric Major Version will ensure compatibility and will continue to work.\n\n## Installation\n#### Quick Start:\n```\npip install ipfabric\n```\n#### Poetry:\n\nIPFabric uses [Poetry](https://pypi.org/project/poetry/) to make setting up a virtual environment with all dependencies\ninstalled quick and easy.\n\nInstall poetry globally:\n```\npip install poetry\n```\nTo install a virtual environment run the following command in the root of this directory.\n\n```\npoetry install\n```\n\nTo run examples, install extras:\n```\npoetry install ipfabric -E examples\n```\n## Introduction\n\nPlease take a look at [API Programmability - Part 1: The Basics](https://ipfabric.io/blog/api-programmability-part-1/)\nfor instructions on creating an API token.\n\nMost of the methods and features can be located in [Examples](examples) to show how to use this package. \nAnother great introduction to this package can be found at [API Programmability - Part 2: Python](https://ipfabric.io/blog/api-programmability-python/)\n\n## Diagrams\n\nDiagramming in IP Fabric version v4.3 and above has been moved to it\'s own package.\n\n```\npip install ipfabric-diagrams\n```\n\n## Authentication\n### Username/Password\nSupply in client:\n```python\nfrom ipfabric import IPFClient\nipf = IPFClient(\'https://demo3.ipfabric.io/\', username=\'user\', password=\'pass\')\n```\n\n### Token\n```python\nfrom ipfabric import IPFClient\nipf = IPFClient(\'https://demo3.ipfabric.io/\', token=\'token\')\n```\n\n### Environment \nThe easiest way to use this package is with a `.env` file.  You can copy the sample and edit it with your environment variables. \n\n```commandline\ncp sample.env .env\n```\n\nThis contains the following variables which can also be set as environment variables instead of a .env file.\n```\nIPF_URL="https://demo3.ipfabric.io"\nIPF_TOKEN=TOKEN\nIPF_VERIFY=true\n```\n\nOr if using Username/Password:\n```\nIPF_URL="https://demo3.ipfabric.io"\nIPF_USERNAME=USER\nIPF_PASSWORD=PASS\n```\n\n**`IPF_DEV` is an internal variable only, do not set to True.**\n\n## Development\nFollow the poetry install then follow instructions below:\n\nTo test and build:\n\n```\npoetry run pytest\npoetry build\n```\n\nPrior to pushing changes run:\n```\npoetry run black ipfabric\npoetry update\npoetry export -f requirements.txt -o requirements.txt --without-hashes\ngit add requirements.txt\n```',
    'author': 'Justin Jeffery',
    'author_email': 'justin.jeffery@ipfabric.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/community-fabric/python-ipfabric',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



from .configuration import DeviceConfigs
from .restore_intents import RestoreIntents
from .sites import UpdateSiteNames
from .vulnerabilities import Vulnerabilities

__all__ = ["DeviceConfigs", "UpdateSiteNames", "Vulnerabilities", "RestoreIntents"]



import logging
from typing import Any, ClassVar, Union

from pydantic.dataclasses import dataclass

from ipfabric.settings.vendor_api_models import (
    AWS,
    Azure,
    CheckPointApiKey,
    CheckPointUserAuth,
    CiscoFMC,
    Merakiv1,
    NSXT,
    SilverPeak,
    Versa,
    Viptela,
)

API_MODELS = Union[
    AWS, Azure, CheckPointApiKey, CheckPointUserAuth, CiscoFMC, Merakiv1, NSXT, SilverPeak, Versa, Viptela
]

logger = logging.getLogger("python-ipfabric")


@dataclass
class VendorAPI:
    client: Any
    _api_url: ClassVar[str] = "settings/vendor-api"

    def get_vendor_apis(self) -> dict:
        """
        Get all vendor apis and sets them in the Authentication.apis
        :return: self.credentials
        """
        res = self.client.get(self._api_url)
        res.raise_for_status()
        return res.json()

    def add_vendor_api(self, api: API_MODELS):
        params = vars(api)
        if isinstance(api, AWS) and api.assumeRole is None:
            params.pop("assumeRole")
        elif isinstance(api, AWS) and api.assumeRole:
            params["assumeRole"] = dict(role=params["assumeRole"])
        res = self.client.post(self._api_url, json=params)
        res.raise_for_status()
        return res.json()

    def delete_vendor_api(self, api_id: Union[dict, str, int]):
        if isinstance(api_id, dict):
            api_id = api_id["id"]
        elif isinstance(api_id, int):
            api_id = str(api_id)
        res = self.client.delete(self._api_url + api_id)
        res.raise_for_status()
        return res.status_code



import logging
from pydantic import BaseModel
from typing import Any
from ipfabric import models

logger = logging.getLogger("python-ipfabric")


class Sdn(BaseModel):
    client: Any

    @property
    def aci_endpoints(self):
        return models.Table(client=self.client, endpoint="tables/aci/endpoints")

    @property
    def aci_vlan(self):
        return models.Table(client=self.client, endpoint="tables/aci/vlan")

    @property
    def aci_vrf(self):
        return models.Table(client=self.client, endpoint="tables/aci/vtep")

    @property
    def aci_dtep(self):
        return models.Table(client=self.client, endpoint="tables/aci/dtep")

    @property
    def vxlan_vtep(self):
        return models.Table(client=self.client, endpoint="tables/vxlan/vtep")

    @property
    def vxlan_peers(self):
        return models.Table(client=self.client, endpoint="tables/vxlan/peers")

    @property
    def vxlan_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/vxlan/interfaces")

    @property
    def vxlan_vni(self):
        return models.Table(client=self.client, endpoint="tables/vxlan/vni")

    @property
    def apic_controllers(self):
        return models.Table(client=self.client, endpoint="tables/apic/controllers")

    @property
    def apic_contexts(self):
        return models.Table(client=self.client, endpoint="tables/apic/contexts")

    @property
    def apic_bridge_domains(self):
        return models.Table(client=self.client, endpoint="tables/apic/bridge-domains")

    @property
    def apic_applications(self):
        return models.Table(client=self.client, endpoint="tables/apic/applications")

    @property
    def apic_endpoint_groups(self):
        return models.Table(client=self.client, endpoint="tables/apic/endpoint-groups")

    @property
    def apic_endpoint_groups_contracts(self):
        return models.Table(client=self.client, endpoint="tables/apic/endpoint-groups/contracts")

    @property
    def apic_contracts(self):
        return models.Table(client=self.client, endpoint="tables/apic/contracts")



import logging
from pydantic import BaseModel
from typing import Any
from ipfabric import models

logger = logging.getLogger("python-ipfabric")


class PortChannels(BaseModel):
    client: Any

    @property
    def inbound_balancing_table(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/port-channel/balance/inbound")

    @property
    def outbound_balancing_table(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/port-channel/balance/outbound")

    @property
    def member_status_table(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/port-channel/member-status")

    @property
    def mlag_switches(self):
        return models.Table(client=self.client, endpoint="tables/platforms/mlag/switches")

    @property
    def mlag_peers(self):
        return models.Table(client=self.client, endpoint="tables/platforms/mlag/peers")

    @property
    def mlag_pairs(self):
        return models.Table(client=self.client, endpoint="tables/platforms/mlag/pairs")

    @property
    def mlag_cisco_vpc(self):
        return models.Table(client=self.client, endpoint="tables/platforms/mlag/vpc")
